# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.pyparsing import *

from yay import nodes

class Actions(object):
    def boxed(self, str, words, tokens):
        return nodes.Boxed(tokens[0])
    def boxed_int(self, str, words, tokens):
        return nodes.Boxed(int(tokens[0]))
    def boxed_octal(self, str, words, tokens):
        return nodes.Boxed(int(tokens[0], 8))
    def concatenation(self, str, words, tokens):
        if len(tokens) == 1:
            return tokens[0]
        return nodes.Concatenation(*tokens)

actions = Actions()

AND = Keyword("and")
OR = Keyword("or")
IN = Keyword("in")
BINOP = oneOf("= != < > <= >=")

identifier = Word(alphanums+"_") | Keyword("@")
arithSign = Word("+-",exact=1)

octNum = Combine(Optional(arithSign) + Suppress("0") + Word(nums)).setParseAction(actions.boxed_octal)
intNum = Combine(Optional(arithSign) + Word(nums)).setParseAction(actions.boxed_int)

expression = Forward()

def function_call_action(s, w, t):
    return nodes.Function(t[0], t[1])
function_identifier = Word(alphanums+"_")
function_call = function_identifier + Group(Suppress("(") + Optional(expression + ZeroOrMore(Suppress(",") + expression)) + Suppress(")"))
function_call.setParseAction(function_call_action)

filterExpression = Forward()

def filter_bin_comparison_action(s, w, t):
    cls = {
        "=": nodes.Equal,
        "!=": nodes.NotEqual,
        "<": nodes.LessThan,
        "<=": nodes.LessThanEqual,
        ">": nodes.GreaterThan,
        ">=": nodes.GreaterThanEqual,
        }[t[0][1]]
    return cls(t[0][0], t[0][2])
filter_bin_comparison = Group(expression + BINOP + expression)
filter_bin_comparison.setParseAction(filter_bin_comparison_action)

filterCondition = (
    filter_bin_comparison |
    ( "(" + filterExpression + ")" )
    )

def filter_expression_action(s, w, t):
    node = t[0]
    for i in range(1, len(t)):
        if t[i] == "and":
            node = nodes.And(node, t[i+1])
        elif t[i] == "or":
            node = nodes.Or(node, t[i+1])
        i += 1
    return node
filterExpression << filterCondition + ZeroOrMore((AND|OR) + filterExpression)
filterExpression.setParseAction(filter_expression_action)


full_list_access = Suppress("[") + filterExpression + Suppress("]")
full_list_access.setParseAction(lambda s, w, t: nodes.Filter(None, t[0]))


def index_access_action(s, w, t):
    return nodes.Access(None, t[0])
listAccess = Suppress("[") + intNum + Suppress("]")
listAccess.setParseAction(index_access_action)


def full_expression_action(s, w, t):
    node = None
    for token in t:
        if not isinstance(token, nodes.Node):
            node = nodes.Access(node, token)
        else:
            token.container = node
            node = token
    return node

fullExpression = identifier + ZeroOrMore(
    full_list_access |
    listAccess |
    Suppress(".") + identifier
    )
fullExpression.setParseAction(full_expression_action)

expression << (
    octNum |
    intNum |
    function_call |
    fullExpression
    )


bracketed_expression = Suppress("${") + expression + Suppress("}")


def ugh(s, w, t):
    if not t or not t[0]:
        return []
    return actions.boxed(s, w, t)
myrol = restOfLine.copy().setParseAction(ugh)

templated_string = ZeroOrMore(
    bracketed_expression |
    SkipTo("${").leaveWhitespace().setParseAction(actions.boxed)
    ) + myrol
templated_string.setParseAction(actions.concatenation)

as_statement = identifier + Suppress("in") + expression

#print as_statement.parseString("foolist[foo.age < bar.maxage] as person")
#print templated_string.parseString("foo bar {foo.ag} foo bar {foo.age} foo baz")[0]
#print templated_string.parseString("{foo.bar.baz}")[0]
#print repr(expression.parseString("foo.bar[foo.age < 12 and foo.badger > 5][0]")[0])
#print fullExpression.parseString("foo.bar")



# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node, Filter

class Lookup(Node):
    """
    I delay a lookup until resolve time
    """
    def __init__(self, root, value):
        super(Lookup, self).__init__(value)
        self.root = root

    def split(self, strng):
        return strng._formatter_field_name_split()

    def join(self, first, rest):
        parts = [first]
        for is_attr, i in rest:
            if is_attr:
                parts.append(".%s" % i)
            else:
                parts.append("[%s]" % i)
        return "".join(parts)

    def semi_resolve(self, context):
        # This is derived from string.py Formatter so lookup language
        # is consistent
        # field name parser is written in C and available as
        #   str._formatter_field_name_split()
        handled = []

        first, rest = self.value.resolve(context)._formatter_field_name_split()
        obj = self.root.get(context, first, None)
        if not obj:
            raise KeyError("Unable to find '%s'" % self.join(first, handled))

        for is_attr, i in rest:
            handled.append((is_attr, i))
            if not is_attr:
                obj = Filter(obj, i).resolve(context)
            else:
                obj = obj.get(context, i)
            if not obj:
                raise KeyError("Unable to find '%s'" % self.join(first, handled))

        return obj

    def get(self, context, idx, default):
        return self.semi_resolve(context).get(context, idx, default)

    def resolve(self, context):
        return self.semi_resolve(context).resolve(context)




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from yay.nodes import *
from yay.context import RootContext

class TestNodeExpressions(unittest.TestCase):

    def test_and(self):
        for a in (True, False):
            for b in (True, False):
                c = And(Boxed(a), Boxed(b))
                self.failUnlessEqual(c.resolve(None), a and b)

    def test_or(self):
        for a in (True, False):
            for b in (True, False):
                c = Or(Boxed(a), Boxed(b))
                self.failUnlessEqual(c.resolve(None), a or b)

    def test_in(self):
        b = ["a", "b", "c", "d"]
        
        c = In(Boxed("a"), Boxed(b))
        self.failUnless(c.resolve(None))

        c = In(Boxed("z"), Boxed(b))
        self.failUnless(not c.resolve(None))

    def test_equals(self):
        a = Boxed(5)
        b = Boxed(5)
        c = Equal(a, b)
        self.failUnless(c.resolve(None))

    def test_not_equals(self):
        a = Boxed(5)
        b = Boxed(4)
        c = NotEqual(a, b)
        self.failUnless(c.resolve(None))

    def test_lt(self):
        a = Boxed(3)
        b = Boxed(4)
        c = LessThan(a, b)
        self.failUnless(c.resolve(None))

    def test_lte(self):
        a = Boxed(3)
        b = Boxed(4)
        c = LessThanEqual(a, b)
        self.failUnless(c.resolve(None))

    def test_gt(self):
        a = Boxed(5)
        b = Boxed(4)
        c = GreaterThan(a, b)
        self.failUnless(c.resolve(None))

    def test_gte(self):
        a = Boxed(5)
        b = Boxed(4)
        c = GreaterThanEqual(a, b)
        self.failUnless(c.resolve(None))

    def test_function(self):
        context = RootContext({})
        context.functions["foo"] = lambda x: x+5
        r = Function("foo", [Boxed(5)])
        self.failUnlessEqual(r.resolve(context), 10)



# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node, Lookup, Boxed
from yay.context import Context

class ForEach(object):

    def __init__(self, root, value, args):
        self.root = root
        self.value = value

        self.lookup = args[1]
        self.alias = args[0].strip()

    def resolve(self, context):
        resolved = []
        for item in self.lookup.semi_resolve(context):
            newctx = Context(context, {self.alias: item})
            resolved.append(self.value.resolve(newctx))
        return resolved





# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import urllib
import StringIO
import os

class IOpener(object):

    def open(self, uri):
        """ Given a uri, return a YAML compatible stream """
        pass


class FileOpener(IOpener):

    scheme = "file://"

    def open(self, uri):
        if uri.startswith("file://"):
            uri = uri[8:]
        return open(uri, "r")


class UrlOpener(IOpener):

    scheme = "http://"

    def open(self, uri):
        return urllib.urlopen(uri)


class MemOpener(IOpener):

    """
    This is purely a testing stub.

    It has a class-shared dictionary called 'data', where keys are uris and
    values are data to return as a StringIO when open() is called. This means
    test cases can exercise the yay extends machinery without the need to create
    temporary files
    """

    scheme = "mem://"
    data = {}

    def open(self, uri):
        return StringIO.StringIO(self.data[uri])

    @classmethod
    def add(cls, uri, data):
        cls.data[uri] = data

    @classmethod
    def reset(cls):
        cls.data = {}


class Openers(object):

    def __init__(self):
        self.openers = {}
        for cls in IOpener.__subclasses__():
            self.openers[cls.scheme] = cls()

    def open(self, uri):
        for scheme, opener in self.openers.iteritems():
            if uri.startswith(scheme):
                return opener.open(uri)

        # Support direct file paths that dont specify a scheme
        if os.path.exists(uri):
            return self.openers["file://"].open(uri)

        raise ValueError("URI '%s' cannot be opened (unsupported scheme or missing local file)" % uri)




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node

class Select(Node):

    def __init__(self, options, key):
        self.options = options
        self.key = key

    def semi_resolve(self, context):
        key = self.key.resolve(context)
        return self.options.get(context, key)

    def resolve(self, context):
        return self.semi_resolve(context).resolve(context)




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from yay.nodes.boxed import Boxed


class TestResolver(unittest.TestCase):
    pass


class Old:
    def test_string(self):
        r = Resolver({
            "foo": "123",
            "baz": "{foo}",
            }).resolve()

        self.failUnlessEqual(r["baz"], "123")

    def test_list(self):
        r = Resolver({
            "foo": "123",
            "baz": [
                "456",
                "{foo}",
                ],
            }).resolve()

        self.failUnlessEqual(r['baz'], ["456", "123"])

    def test_nested_dict(self):
        r = Resolver({
            "foo": 123,
            "baz": {
                "foo": "{foo}",
                },
            }).resolve()

        self.failUnlessEqual(r['baz']['foo'], "123")

    def test_cycle_detection(self):
        r = Resolver({
            "foo": "{baz}",
            "baz": "{foo}",
            })

        self.failUnlessRaises(ValueError, r.resolve)




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from yay.parser import templated_string, expression

class TestTemplatedString(unittest.TestCase):

    def do(self, teststring, expect):
        self.failUnlessEqual(
            repr(templated_string.parseString(teststring)[0]),
            expect
            )

    def test_template_whole_string(self):
        self.do("${foo}", "Access(None, foo)")

    def test_template_middle_of_string(self):
        self.do("foo${bar}baz", "Concat(Boxed(foo), Access(None, bar), Boxed(baz))")

    def test_template_middle_of_string_with_spaces(self):
        self.do("foo ${bar} baz", "Concat(Boxed(foo ), Access(None, bar), Boxed( baz))")

    def test_template_start_of_string(self):
        self.do("${foo} bar", "Concat(Access(None, foo), Boxed( bar))")

    def test_template_end_of_string(self):
        self.do("foo ${bar}", "Concat(Boxed(foo ), Access(None, bar))")

class TestFunctionCall(unittest.TestCase):

    def do(self, teststring, expect):
        self.failUnlessEqual(
            repr(expression.parseString(teststring)[0]),
            expect
            )

    def test_parse_simple(self):
        self.do("range(4, 5)", "Function(range, Boxed(4), Boxed(5))")

    def test_parse_nested(self):
        self.do("range(min(4, 5), 5)", "Function(range, Function(min, Boxed(4), Boxed(5)), Boxed(5))")




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node

class Append(Node):

    def get(self, context, idx, default=None):
        return Boxed(self.resolve(context)[int(idx)])

    def apply(self, context, existing):
        if not existing:
            existing = []
        return existing + self.value.resolve(context)




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Node(object):
    __slots__ = ("chain", "value")
    chain = None

    def __init__(self, value=None):
        # Premature typing optimisation
        self.value = value

    def apply(self, context, data):
        pass

    def resolve(self, context):
        data = None
        if self.chain:
            data = self.chain.resolve(context)
        return self.apply(context, data)

    def __str__(self):
        return repr(self)



# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yaml.error import MarkedYAMLError
from yaml.events import ScalarEvent, SequenceStartEvent, SequenceEndEvent, \
    MappingStartEvent, MappingEndEvent, AliasEvent, StreamEndEvent

from yay.nodes import *
from yay.parser import templated_string, as_statement, expression


class ComposerError(MarkedYAMLError):
    pass


class Composer(object):

    def __init__(self):
        self.root = None
        self.action_map = {
            "copy": lambda value, args: Copy(value),
            "assign": lambda value, args: value if isinstance(value, Node) else Boxed(value),
            "append": lambda value, args: Append(value),
            "remove": lambda value, args: Remove(value),
            "foreach": lambda value, args: ForEach(self, value, as_statement.parseString(args)),
            "select": lambda value, args: Select(value, expression.parseString(args)[0]),
            "flatten": lambda value, args: Flatten(value),
            }
        self.dirty = False

    def compose(self, previous):
        # Drop the STREAM-START event.
        self.get_event()

        # Compose a document if the stream is not empty.
        document = None
        if not self.check_event(StreamEndEvent):
            self.get_event() # Drop DOCUMENT-START
            document = self.compose_node(previous)
            self.get_event() # Drop DOCUMENT-END

        # Ensure that the stream contains no more documents.
        if not self.check_event(StreamEndEvent):
            event = self.get_event()
            raise ComposerError("expected a single document in the stream",
                    document.start_mark, "but found another document",
                    event.start_mark)

        # Drop the STREAM-END event.
        self.get_event()

        return document

    def compose_node(self, previous):
        if self.check_event(AliasEvent):
            raise ComposerError(None, None, "found alias, these arent supported in yay", event.start_mark)

        node = None

        if self.check_event(ScalarEvent):
            node = self.compose_scalar(previous)
        elif self.check_event(SequenceStartEvent):
            node = self.compose_sequence(previous)
        elif self.check_event(MappingStartEvent):
            node = self.compose_mapping_or_anonymous(previous)

        if not node:
            event = self.peek_event()
            raise ComposerError(None, None, "unexpected event in stream", event.start_mark)

        return node

    def compose_scalar(self, previous):
        event = self.get_event()

        if isinstance(event.value, basestring):
            #Icky - this needs to move *beneath* this layer of code
            node = templated_string.parseString(event.value)[0]
        else:
            node = Boxed(event.value)

        node.start_mark = event.start_mark
        node.end_mark = event.end_mark

        return node

    def compose_sequence(self, previous):
        start = self.get_event()

        data = []
        while not self.check_event(SequenceEndEvent):
            data.append(self.compose_node(None))

        end = self.get_event()

        node = Sequence(data)
        node.start_mark = start.start_mark
        node.end_mark = end.end_mark

        return node

    def handle_special_term(self, previous):
        key_event = self.peek_event()
        if key_event.value != self.special_term:
            return previous
        self.get_event()

        special_term = self.compose_node(None).resolve(None)

        for extend in special_term.get("extends", []):
            data = self.openers.open(extend)
            previous = self.__class__(data, special_term=self.special_term).compose(previous)

        return previous

    def compose_mapping_or_anonymous(self, previous):
        start = self.get_event()

        if self.peek_event().value.startswith("."):
            value = self.compose_anonymous(previous)
        else:
            value = self.compose_mapping(previous)

        end = self.get_event()

        value.start_mark = start.start_mark
        value.end_mark = end.end_mark

        return value

    def compose_anonymous(self, previous):
        # An anonymous expression that doesnt bind to a keyword
        key_event = self.get_event()
        action = key_event.value[1:]

        action_args = None
        if " " in action:
            action, action_args = action.split(" ", 1)

        value = self.compose_node(None)

        assert self.check_event(MappingEndEvent)

        return self.action_map[action](value, action_args)

    def compose_mapping(self, previous):
        if not self.dirty:
            previous = self.handle_special_term(previous)
            self.dirty = True

        container = Mapping(previous)
        while not self.check_event(MappingEndEvent):
            key_event = self.get_event()
            key = key_event.value

            action = "assign"
            if "." in key:
                key, action = key.split(".", 1)

            action_args = None
            if " " in action:
                action, action_args = action.split(" ", 1)

            # FIXME: context-driven traversal is different from non-resolving dict-lookup
            existing = container.get(None, key, None)

            # Grab scalar value
            boxed = self.compose_node(existing)

            # Further box the value based on the kind of action it is
            boxed = self.action_map[action](boxed, action_args)

            # Make sure that Appends are hooked up to correct List
            boxed.chain = existing

            # And add it to the dictionary (which will automatically chain nodes)
            container.set(key, boxed)

        return container




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml

from yay.loader import Loader
from yay.openers import Openers
from yay.context import RootContext

class Config(object):

    def __init__(self, special_term='yay'):
        self.special_term = special_term
        self.openers = Openers()
        self.clear()

    def load_uri(self, uri):
        self.load(self.openers.open(uri))

    def load(self, stream):
        data = Loader(stream, special_term=self.special_term).compose(self.mapping)
        self.mapping = data

    def clear(self):
        self.mapping = None

    def get(self):
        if not self.mapping:
            return {}
        return self.mapping.resolve(RootContext(self.mapping))

def load_uri(uri, special_term='yay'):
    c = Config(special_term)
    c.load_uri(uri)
    return c.get()

def load(stream, special_term='yay'):
    c = Config(special_term)
    c.load(stream)
    return c.get()

